@@ -1,28 +1,18 @@
 import sys
-from attr import (
-    Factory,
-    NOTHING,
-    fields as attrs_fields,
-    Attribute,
-)
-from dataclasses import (
-    MISSING,
-    is_dataclass,
-    fields as dataclass_fields,
-)
-from typing import (
-    Any,
-    Dict,
-    FrozenSet,
-    List,
-    Mapping,
-    MutableMapping,
-    MutableSequence,
-    MutableSet,
-    Sequence,
-    Set,
-    Tuple,
-)
+from dataclasses import MISSING
+from dataclasses import fields as dataclass_fields
+from dataclasses import is_dataclass
+from typing import Any, Dict, FrozenSet, List
+from typing import Mapping as TypingMapping
+from typing import MutableMapping as TypingMutableMapping
+from typing import MutableSequence as TypingMutableSequence
+from typing import MutableSet as TypingMutableSet
+from typing import Sequence as TypingSequence
+from typing import Set as TypingSet
+from typing import Tuple
+
+from attr import NOTHING, Attribute, Factory
+from attr import fields as attrs_fields
 
 version_info = sys.version_info[0:3]
 is_py37 = version_info[:2] == (3, 7)
@@ -48,6 +38,11 @@ def has(cls):
     )
 
 
+def has_with_generic(cls):
+    """Test whether the class if a normal or generic attrs or dataclass."""
+    return has(cls) or has(get_origin(cls))
+
+
 def fields(type):
     try:
         return type.__attrs_attrs__
@@ -85,8 +80,23 @@ def adapted_fields(type) -> List[Attribute]:
         return attrs_fields(type)
 
 
+def is_hetero_tuple(type: Any) -> bool:
+    origin = getattr(type, "__origin__", None)
+    return origin is tuple and ... not in type.__args__
+
+
 if is_py37 or is_py38:
-    from typing import Union, _GenericAlias
+    Set = TypingSet
+    MutableSet = TypingMutableSet
+    Sequence = TypingSequence
+    MutableSequence = TypingMutableSequence
+    MutableMapping = TypingMutableMapping
+    Mapping = TypingMapping
+    FrozenSetSubscriptable = FrozenSet
+    TupleSubscriptable = Tuple
+
+    from collections import Counter as ColCounter
+    from typing import Counter, Union, _GenericAlias
 
     def is_annotated(_):
         return False
@@ -107,8 +117,11 @@ if is_py37 or is_py38:
     def is_sequence(type: Any) -> bool:
         return type in (List, list, Tuple, tuple) or (
             type.__class__ is _GenericAlias
-            and type.__origin__ is not Union
-            and issubclass(type.__origin__, Sequence)
+            and (
+                type.__origin__ not in (Union, Tuple, tuple)
+                and issubclass(type.__origin__, TypingSequence)
+            )
+            or (type.__origin__ in (Tuple, tuple) and type.__args__[1] is ...)
         )
 
     def is_mutable_set(type):
@@ -124,16 +137,16 @@ if is_py37 or is_py38:
         )
 
     def is_mapping(type):
-        return type in (Mapping, dict) or (
+        return type in (TypingMapping, dict) or (
             type.__class__ is _GenericAlias
-            and issubclass(type.__origin__, Mapping)
+            and issubclass(type.__origin__, TypingMapping)
         )
 
     bare_list_args = List.__args__
-    bare_seq_args = Sequence.__args__
-    bare_mapping_args = Mapping.__args__
+    bare_seq_args = TypingSequence.__args__
+    bare_mapping_args = TypingMapping.__args__
     bare_dict_args = Dict.__args__
-    bare_mutable_seq_args = MutableSequence.__args__
+    bare_mutable_seq_args = TypingMutableSequence.__args__
 
     def is_bare(type):
         args = type.__args__
@@ -145,23 +158,65 @@ if is_py37 or is_py38:
             or args == bare_mutable_seq_args
         )
 
+    def is_counter(type):
+        return (
+            type in (Counter, ColCounter)
+            or getattr(type, "__origin__", None) is ColCounter
+        )
+
+    if is_py38:
+        from typing import Literal
+
+        def is_literal(type) -> bool:
+            return (
+                type.__class__ is _GenericAlias and type.__origin__ is Literal
+            )
+
+    else:
+        # No literals in 3.7.
+        def is_literal(_) -> bool:
+            return False
+
 
 else:
     # 3.9+
+    from collections import Counter
+    from collections.abc import Mapping as AbcMapping
+    from collections.abc import MutableMapping as AbcMutableMapping
+    from collections.abc import MutableSequence as AbcMutableSequence
+    from collections.abc import MutableSet as AbcMutableSet
+    from collections.abc import Sequence as AbcSequence
+    from collections.abc import Set as AbcSet
+    from typing import Counter as TypingCounter
     from typing import (
         Union,
+        _AnnotatedAlias,
         _GenericAlias,
         _SpecialGenericAlias,
         _UnionGenericAlias,
-        _AnnotatedAlias,
-    )
-    from collections.abc import (
-        MutableSequence as AbcMutableSequence,
-        Sequence as AbcSequence,
-        MutableSet as AbcMutableSet,
-        Set as AbcSet,
     )
 
+    try:
+        # Not present on 3.9.0, so we try carefully.
+        from typing import _LiteralGenericAlias
+
+        def is_literal(type) -> bool:
+            return type.__class__ is _LiteralGenericAlias
+
+    except ImportError:
+
+        def is_literal(_) -> bool:
+            return False
+
+    Set = AbcSet
+    MutableSet = AbcMutableSet
+    Sequence = AbcSequence
+    MutableSequence = AbcMutableSequence
+    MutableMapping = AbcMutableMapping
+    Mapping = AbcMapping
+    FrozenSetSubscriptable = frozenset
+    TupleSubscriptable = tuple
+
     def is_annotated(type) -> bool:
         return getattr(type, "__class__", None) is _AnnotatedAlias
 
@@ -189,8 +244,8 @@ else:
             in (
                 List,
                 list,
-                Sequence,
-                MutableSequence,
+                TypingSequence,
+                TypingMutableSequence,
                 AbcMutableSequence,
                 tuple,
             )
@@ -198,10 +253,7 @@ else:
                 type.__class__ is _GenericAlias
                 and (
                     (origin is not tuple)
-                    and issubclass(
-                        origin,
-                        Sequence,
-                    )
+                    and issubclass(origin, TypingSequence)
                     or origin is tuple
                     and type.__args__[1] is ...
                 )
@@ -212,10 +264,10 @@ else:
 
     def is_mutable_set(type):
         return (
-            type in (Set, MutableSet, set)
+            type in (TypingSet, TypingMutableSet, set)
             or (
                 type.__class__ is _GenericAlias
-                and issubclass(type.__origin__, MutableSet)
+                and issubclass(type.__origin__, TypingMutableSet)
             )
             or (
                 getattr(type, "__origin__", None)
@@ -240,14 +292,35 @@ else:
 
     def is_mapping(type):
         return (
-            type in (Mapping, Dict, MutableMapping, dict)
+            type
+            in (
+                TypingMapping,
+                Dict,
+                TypingMutableMapping,
+                dict,
+                AbcMutableMapping,
+            )
             or (
                 type.__class__ is _GenericAlias
-                and issubclass(type.__origin__, Mapping)
+                and issubclass(type.__origin__, TypingMapping)
             )
-            or (getattr(type, "__origin__", None) is dict)
+            or (
+                getattr(type, "__origin__", None)
+                in (dict, AbcMutableMapping, AbcMapping)
+            )
+            or issubclass(type, dict)
+        )
+
+    def is_counter(type):
+        return (
+            type in (Counter, TypingCounter)
+            or getattr(type, "__origin__", None) is Counter
         )
 
 
 def is_generic(obj):
     return isinstance(obj, _GenericAlias)
+
+
+def is_generic_attrs(type):
+    return is_generic(type) and has(type.__origin__)
