@@ -1,30 +1,50 @@
+from collections import Counter
+from collections.abc import MutableSet as AbcMutableSet
+from dataclasses import Field
 from enum import Enum
 from functools import lru_cache
-from typing import Any, Callable, Dict, Mapping, Optional, Tuple, Type, TypeVar
+from typing import Any, Callable, Dict, Optional, Tuple, Type, TypeVar, Union
 
-from attr import resolve_types, has as attrs_has
+from attr import Attribute
+from attr import has as attrs_has
+from attr import resolve_types
 
 from ._compat import (
+    FrozenSetSubscriptable,
+    Mapping,
+    MutableMapping,
+    MutableSequence,
+    MutableSet,
+    Sequence,
+    Set,
+    fields,
     get_origin,
+    has,
+    has_with_generic,
+    is_annotated,
     is_bare,
+    is_counter,
     is_frozenset,
-    is_generic,
+    is_generic_attrs,
+    is_hetero_tuple,
+    is_literal,
     is_mapping,
     is_mutable_set,
     is_sequence,
     is_tuple,
     is_union_type,
-    is_annotated,
-    has,
-    fields,
 )
 from .disambiguators import create_uniq_field_dis_func
 from .dispatch import MultiStrategyDispatch
+from .errors import StructureHandlerNotFoundError
 from .gen import (
     AttributeOverride,
     make_dict_structure_fn,
     make_dict_unstructure_fn,
+    make_hetero_tuple_unstructure_fn,
     make_iterable_unstructure_fn,
+    make_mapping_structure_fn,
+    make_mapping_unstructure_fn,
 )
 
 NoneType = type(None)
@@ -40,10 +60,30 @@ class UnstructureStrategy(Enum):
 
 
 def _subclass(typ):
-    """ a shortcut """
+    """a shortcut"""
     return lambda cls: issubclass(cls, typ)
 
 
+def is_attrs_union(typ):
+    return is_union_type(typ) and all(
+        has(get_origin(e) or e) for e in typ.__args__
+    )
+
+
+def is_attrs_union_or_none(typ):
+    return is_union_type(typ) and all(
+        e is NoneType or has(get_origin(e) or e) for e in typ.__args__
+    )
+
+
+def is_optional(typ):
+    return (
+        is_union_type(typ)
+        and NoneType in typ.__args__
+        and len(typ.__args__) == 2
+    )
+
+
 class Converter(object):
     """Converts between structured and unstructured data."""
 
@@ -55,14 +95,17 @@ class Converter(object):
         "_dict_factory",
         "_union_struct_registry",
         "_structure_func",
+        "_prefer_attrib_converters",
     )
 
     def __init__(
         self,
         dict_factory: Callable[[], Any] = dict,
         unstruct_strat: UnstructureStrategy = UnstructureStrategy.AS_DICT,
+        prefer_attrib_converters: bool = False,
     ) -> None:
         unstruct_strat = UnstructureStrategy(unstruct_strat)
+        self._prefer_attrib_converters = prefer_attrib_converters
 
         # Create a per-instance cache.
         if unstruct_strat is UnstructureStrategy.AS_DICT:
@@ -98,25 +141,38 @@ class Converter(object):
         # Per-instance register of to-attrs converters.
         # Singledispatch dispatches based on the first argument, so we
         # store the function and switch the arguments in self.loads.
-        self._structure_func = MultiStrategyDispatch(self._structure_default)
+        self._structure_func = MultiStrategyDispatch(self._structure_error)
         self._structure_func.register_func_list(
             [
+                (
+                    lambda cl: cl is Any or cl is Optional or cl is None,
+                    lambda v, _: v,
+                ),
+                (is_generic_attrs, self._gen_structure_generic, True),
+                (is_literal, self._structure_literal),
                 (is_sequence, self._structure_list),
                 (is_mutable_set, self._structure_set),
                 (is_frozenset, self._structure_frozenset),
                 (is_tuple, self._structure_tuple),
                 (is_mapping, self._structure_dict),
-                (is_union_type, self._structure_union),
+                (
+                    is_attrs_union_or_none,
+                    self._gen_attrs_union_structure,
+                    True,
+                ),
+                (
+                    lambda t: is_union_type(t)
+                    and t in self._union_struct_registry,
+                    self._structure_union,
+                ),
+                (is_optional, self._structure_optional),
                 (has, self._structure_attrs),
             ]
         )
         # Strings are sequences.
         self._structure_func.register_cls_list(
             [
-                (
-                    str,
-                    self._structure_call,
-                ),
+                (str, self._structure_call),
                 (bytes, self._structure_call),
                 (int, self._structure_call),
                 (float, self._structure_call),
@@ -170,6 +226,22 @@ class Converter(object):
         """
         self._unstructure_func.register_func_list([(check_func, func)])
 
+    def register_unstructure_hook_factory(
+        self,
+        predicate: Callable[[Any], bool],
+        factory: Callable[[Any], Callable[[Any], Any]],
+    ) -> None:
+        """
+        Register a hook factory for a given predicate.
+
+        A predicate is a function that, given a type, returns whether the factory
+        can produce a hook for that type.
+
+        A factory is a callable that, given a type, produces an unstructuring
+        hook for that type. This unstructuring hook will be cached.
+        """
+        self._unstructure_func.register_func_list([(predicate, factory, True)])
+
     def register_structure_hook(
         self, cl: Any, func: Callable[[Any, Type[T]], T]
     ):
@@ -186,6 +258,7 @@ class Converter(object):
             resolve_types(cl)
         if is_union_type(cl):
             self._union_struct_registry[cl] = func
+            self._structure_func.clear_cache()
         else:
             self._structure_func.register_cls_list([(cl, func)])
 
@@ -199,6 +272,22 @@ class Converter(object):
         """
         self._structure_func.register_func_list([(check_func, func)])
 
+    def register_structure_hook_factory(
+        self,
+        predicate: Callable[[Any], bool],
+        factory: Callable[[Any], Callable[[Any], Any]],
+    ) -> None:
+        """
+        Register a hook factory for a given predicate.
+
+        A predicate is a function that, given a type, returns whether the factory
+        can produce a hook for that type.
+
+        A factory is a callable that, given a type, produces a structuring
+        hook for that type. This structuring hook will be cached.
+        """
+        self._structure_func.register_func_list([(predicate, factory, True)])
+
     def structure(self, obj: Any, cl: Type[T]) -> T:
         """Convert unstructured Python data structures to structured data."""
 
@@ -262,28 +351,41 @@ class Converter(object):
 
     # Python primitives to classes.
 
-    def _structure_default(self, obj, cl):
-        """This is the fallthrough case. Everything is a subclass of `Any`.
-
-        A special condition here handles ``attrs`` classes.
-
-        Bare optionals end here too (optionals with arguments are unions.) We
-        treat bare optionals as Any.
-        """
-        if cl is Any or cl is Optional or cl is None:
-            return obj
-
-        if is_generic(cl):
-            fn = make_dict_structure_fn(cl, self)
-            self.register_structure_hook(cl, fn)
-            return fn(obj)
-
-        # We don't know what this is, so we complain loudly.
+    def _structure_error(self, _, cl):
+        """At the bottom of the condition stack, we explode if we can't handle it."""
         msg = (
             "Unsupported type: {0}. Register a structure hook for "
             "it.".format(cl)
         )
-        raise ValueError(msg)
+        raise StructureHandlerNotFoundError(msg, type_=cl)
+
+    def _gen_structure_generic(self, cl):
+        """Create and return a hook for structuring generics."""
+        fn = make_dict_structure_fn(
+            cl,
+            self,
+            _cattrs_prefer_attrib_converters=self._prefer_attrib_converters,
+        )
+        return fn
+
+    def _gen_attrs_union_structure(self, cl):
+        """Generate a structuring function for a union of attrs classes (and maybe None)."""
+        dis_fn = self._get_dis_func(cl)
+        has_none = NoneType in cl.__args__
+
+        if has_none:
+
+            def structure_attrs_union(obj, _):
+                if obj is None:
+                    return None
+                return self.structure(obj, dis_fn(obj))
+
+        else:
+
+            def structure_attrs_union(obj, _):
+                return self.structure(obj, dis_fn(obj))
+
+        return structure_attrs_union
 
     @staticmethod
     def _structure_call(obj, cl):
@@ -295,6 +397,12 @@ class Converter(object):
         """
         return cl(obj)
 
+    @staticmethod
+    def _structure_literal(val, type):
+        if val != type.__args__[0]:
+            raise Exception(f"Literal {type} not equal to {val}")
+        return val
+
     # Attrs classes.
 
     def structure_attrs_fromtuple(
@@ -304,18 +412,34 @@ class Converter(object):
         conv_obj = []  # A list of converter parameters.
         for a, value in zip(fields(cl), obj):  # type: ignore
             # We detect the type by the metadata.
-            converted = self._structure_attr_from_tuple(a, a.name, value)
+            converted = self._structure_attribute(a, value)
             conv_obj.append(converted)
 
         return cl(*conv_obj)  # type: ignore
 
-    def _structure_attr_from_tuple(self, a, _, value):
+    def _structure_attribute(
+        self, a: Union[Attribute, Field], value: Any
+    ) -> Any:
         """Handle an individual attrs attribute."""
         type_ = a.type
+        attrib_converter = getattr(a, "converter", None)
+        if self._prefer_attrib_converters and attrib_converter:
+            # A attrib converter is defined on this attribute, and prefer_attrib_converters is set
+            # to give these priority over registered structure hooks. So, pass through the raw
+            # value, which attrs will flow into the converter
+            return value
         if type_ is None:
             # No type metadata.
             return value
-        return self._structure_func.dispatch(type_)(value, type_)
+
+        try:
+            return self._structure_func.dispatch(type_)(value, type_)
+        except StructureHandlerNotFoundError:
+            if attrib_converter:
+                # Return the original value and fallback to using an attrib converter.
+                return value
+            else:
+                raise
 
     def structure_attrs_fromdict(
         self, obj: Mapping[str, Any], cl: Type[T]
@@ -324,10 +448,7 @@ class Converter(object):
         # For public use.
 
         conv_obj = {}  # Start with a fresh dict, to ignore extra keys.
-        dispatch = self._structure_func.dispatch
         for a in fields(cl):  # type: ignore
-            # We detect the type by metadata.
-            type_ = a.type
             name = a.name
 
             try:
@@ -338,9 +459,7 @@ class Converter(object):
             if name[0] == "_":
                 name = name[1:]
 
-            conv_obj[name] = (
-                dispatch(type_)(val, type_) if type_ is not None else val
-            )
+            conv_obj[name] = self._structure_attribute(a, val)
 
         return cl(**conv_obj)  # type: ignore
 
@@ -395,35 +514,20 @@ class Converter(object):
                     for k, v in obj.items()
                 }
 
-    def _structure_union(self, obj, union):
-        """Deal with converting a union."""
-        # Unions with NoneType in them are basically optionals.
-        # We check for NoneType early and handle the case of obj being None,
-        # so disambiguation functions don't need to handle NoneType.
+    def _structure_optional(self, obj, union):
+        if obj is None:
+            return None
         union_params = union.__args__
-        if NoneType in union_params:  # type: ignore
-            if obj is None:
-                return None
-            if len(union_params) == 2:
-                # This is just a NoneType and something else.
-                other = (
-                    union_params[0]
-                    if union_params[1] is NoneType  # type: ignore
-                    else union_params[1]
-                )
-                # We can't actually have a Union of a Union, so this is safe.
-                return self._structure_func.dispatch(other)(obj, other)
+        other = (
+            union_params[0] if union_params[1] is NoneType else union_params[1]
+        )
+        # We can't actually have a Union of a Union, so this is safe.
+        return self._structure_func.dispatch(other)(obj, other)
 
-        # Check the union registry first.
-        handler = self._union_struct_registry.get(union)
-        if handler is not None:
-            return handler(obj, union)
-
-        # Getting here means either this is not an optional, or it's an
-        # optional with more than one parameter.
-        # Let's support only unions of attr classes for now.
-        cl = self._dis_func_cache(union)(obj)
-        return self._structure_func.dispatch(cl)(obj, cl)
+    def _structure_union(self, obj, union):
+        """Deal with structuring a union."""
+        handler = self._union_struct_registry[union]
+        return handler(obj, union)
 
     def _structure_tuple(self, obj, tup: Type[T]):
         """Deal with converting to a tuple."""
@@ -460,9 +564,10 @@ class Converter(object):
             )
 
         if not all(has(get_origin(e) or e) for e in union_types):
-            raise ValueError(
-                "Only unions of attr classes supported "
-                "currently. Register a loads hook manually."
+            raise StructureHandlerNotFoundError(
+                "Only unions of attrs classes supported "
+                "currently. Register a loads hook manually.",
+                type_=union,
             )
         return create_uniq_field_dis_func(*union_types)
 
@@ -470,64 +575,120 @@ class Converter(object):
 class GenConverter(Converter):
     """A converter which generates specialized un/structuring functions."""
 
-    __slots__ = ("omit_if_default", "type_overrides")
+    __slots__ = (
+        "omit_if_default",
+        "forbid_extra_keys",
+        "type_overrides",
+        "_unstruct_collection_overrides",
+    )
 
     def __init__(
         self,
         dict_factory: Callable[[], Any] = dict,
         unstruct_strat: UnstructureStrategy = UnstructureStrategy.AS_DICT,
         omit_if_default: bool = False,
+        forbid_extra_keys: bool = False,
         type_overrides: Mapping[Type, AttributeOverride] = {},
+        unstruct_collection_overrides: Mapping[Type, Callable] = {},
+        prefer_attrib_converters: bool = False,
     ):
         super().__init__(
-            dict_factory=dict_factory, unstruct_strat=unstruct_strat
+            dict_factory=dict_factory,
+            unstruct_strat=unstruct_strat,
+            prefer_attrib_converters=prefer_attrib_converters,
         )
         self.omit_if_default = omit_if_default
-        self.type_overrides = type_overrides
+        self.forbid_extra_keys = forbid_extra_keys
+        self.type_overrides = dict(type_overrides)
+
+        self._unstruct_collection_overrides = unstruct_collection_overrides
+
+        # Do a little post-processing magic to make things easier for users.
+        co = unstruct_collection_overrides
+
+        # abc.Set overrides, if defined, apply to abc.MutableSets and sets
+        if Set in co:
+            if MutableSet not in co:
+                co[MutableSet] = co[Set]
+                co[AbcMutableSet] = co[Set]  # For 3.7/3.8 compatibility.
+            if FrozenSetSubscriptable not in co:
+                co[FrozenSetSubscriptable] = co[Set]
+
+        # abc.MutableSet overrrides, if defined, apply to sets
+        if MutableSet in co:
+            if set not in co:
+                co[set] = co[MutableSet]
+
+        if FrozenSetSubscriptable in co:
+            co[frozenset] = co[
+                FrozenSetSubscriptable
+            ]  # For 3.7/3.8 compatibility.
+
+        # abc.Sequence overrides, if defined, can apply to MutableSequences, lists and tuples
+        if Sequence in co:
+            if MutableSequence not in co:
+                co[MutableSequence] = co[Sequence]
+            if tuple not in co:
+                co[tuple] = co[Sequence]
+
+        # abc.MutableSequence overrides, if defined, can apply to lists
+        if MutableSequence in co:
+            if list not in co:
+                co[list] = co[MutableSequence]
+
+        # abc.Mapping overrides, if defined, can apply to MutableMappings
+        if Mapping in co:
+            if MutableMapping not in co:
+                co[MutableMapping] = co[Mapping]
+
+        # abc.MutableMapping overrides, if defined, can apply to dicts
+        if MutableMapping in co:
+            if dict not in co:
+                co[dict] = co[MutableMapping]
+
+        # builtins.dict overrides, if defined, can apply to counters
+        if dict in co:
+            if Counter not in co:
+                co[Counter] = co[dict]
 
         if unstruct_strat is UnstructureStrategy.AS_DICT:
             # Override the attrs handler.
-            self._unstructure_func.register_func_list(
-                [
-                    (
-                        has,
-                        self.gen_unstructure_attrs_fromdict,
-                        True,
-                    ),
-                ]
+            self.register_unstructure_hook_factory(
+                has_with_generic, self.gen_unstructure_attrs_fromdict
             )
-            self._structure_func.register_func_list(
-                [
-                    (has, self.gen_structure_attrs_fromdict, True),
-                ]
+            self.register_structure_hook_factory(
+                has, self.gen_structure_attrs_fromdict
             )
-
-        self._unstructure_func.register_func_list(
-            [
-                (is_annotated, self.gen_unstructure_annotated, True),
-                (
-                    is_sequence,
-                    self.gen_unstructure_iterable,
-                    True,
-                ),
-                (
-                    is_mutable_set,
-                    lambda cl: self.gen_unstructure_iterable(
-                        cl, unstructure_to=set
-                    ),
-                    True,
-                ),
-                (
-                    is_frozenset,
-                    lambda cl: self.gen_unstructure_iterable(
-                        cl, unstructure_to=frozenset
-                    ),
-                    True,
-                ),
-            ]
+        self.register_unstructure_hook_factory(
+            is_annotated, self.gen_unstructure_annotated
         )
-        self._structure_func.register_func_list(
-            [(is_annotated, self.gen_structure_annotated, True)]
+        self.register_unstructure_hook_factory(
+            is_hetero_tuple, self.gen_unstructure_hetero_tuple
+        )
+        self.register_unstructure_hook_factory(
+            is_sequence, self.gen_unstructure_iterable
+        )
+        self.register_unstructure_hook_factory(
+            is_mapping, self.gen_unstructure_mapping
+        )
+        self.register_unstructure_hook_factory(
+            is_mutable_set,
+            lambda cl: self.gen_unstructure_iterable(cl, unstructure_to=set),
+        )
+        self.register_unstructure_hook_factory(
+            is_frozenset,
+            lambda cl: self.gen_unstructure_iterable(
+                cl, unstructure_to=frozenset
+            ),
+        )
+        self.register_structure_hook_factory(
+            is_annotated, self.gen_structure_annotated
+        )
+        self.register_structure_hook_factory(
+            is_mapping, self.gen_structure_mapping
+        )
+        self.register_structure_hook_factory(
+            is_counter, self.gen_structure_counter
         )
 
     def gen_unstructure_annotated(self, type):
@@ -541,6 +702,9 @@ class GenConverter(Converter):
         return h
 
     def gen_unstructure_attrs_fromdict(self, cl: Type[T]) -> Dict[str, Any]:
+        origin = get_origin(cl)
+        if origin is not None:
+            cl = origin
         attribs = fields(cl)
         if any(isinstance(a.type, str) for a in attribs):
             # PEP 563 annotations - need to be resolved.
@@ -554,7 +718,6 @@ class GenConverter(Converter):
         h = make_dict_unstructure_fn(
             cl, self, omit_if_default=self.omit_if_default, **attrib_overrides
         )
-        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
         return h
 
     def gen_structure_attrs_fromdict(self, cl: Type[T]) -> T:
@@ -567,14 +730,56 @@ class GenConverter(Converter):
             for a in attribs
             if a.type in self.type_overrides
         }
-        h = make_dict_structure_fn(cl, self, **attrib_overrides)
-        self._structure_func.register_cls_list([(cl, h)], direct=True)
+        h = make_dict_structure_fn(
+            cl,
+            self,
+            _cattrs_forbid_extra_keys=self.forbid_extra_keys,
+            _cattrs_prefer_attrib_converters=self._prefer_attrib_converters,
+            **attrib_overrides,
+        )
         # only direct dispatch so that subclasses get separately generated
         return h
 
-    def gen_unstructure_iterable(self, cl: Any, unstructure_to=list):
+    def gen_unstructure_iterable(self, cl: Any, unstructure_to=None):
+        unstructure_to = self._unstruct_collection_overrides.get(
+            get_origin(cl) or cl, unstructure_to or list
+        )
         h = make_iterable_unstructure_fn(
             cl, self, unstructure_to=unstructure_to
         )
         self._unstructure_func.register_cls_list([(cl, h)], direct=True)
         return h
+
+    def gen_unstructure_hetero_tuple(self, cl: Any, unstructure_to=None):
+        unstructure_to = self._unstruct_collection_overrides.get(
+            get_origin(cl) or cl, unstructure_to or tuple
+        )
+        h = make_hetero_tuple_unstructure_fn(
+            cl, self, unstructure_to=unstructure_to
+        )
+        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
+        return h
+
+    def gen_unstructure_mapping(
+        self, cl: Any, unstructure_to=None, key_handler=None
+    ):
+        unstructure_to = self._unstruct_collection_overrides.get(
+            get_origin(cl) or cl, unstructure_to or dict
+        )
+        h = make_mapping_unstructure_fn(
+            cl, self, unstructure_to=unstructure_to, key_handler=key_handler
+        )
+        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
+        return h
+
+    def gen_structure_counter(self, cl: Any):
+        h = make_mapping_structure_fn(
+            cl, self, structure_to=Counter, val_type=int
+        )
+        self._structure_func.register_cls_list([(cl, h)], direct=True)
+        return h
+
+    def gen_structure_mapping(self, cl: Any):
+        h = make_mapping_structure_fn(cl, self)
+        self._structure_func.register_cls_list([(cl, h)], direct=True)
+        return h
